import numpy as np
import pandas as pd

import artistools as at

modelpath = at.get_config()["path_testartismodel"]
modelpath_3d = at.get_config()["path_testartismodel"].parent / "testmodel_3d_10^3"
outputpath = at.get_config()["path_testoutput"]


def test_describeinputmodel() -> None:
    at.inputmodel.describeinputmodel.main(argsraw=[], inputfile=modelpath, get_elemabundances=True)


def test_describeinputmodel_3d() -> None:
    at.inputmodel.describeinputmodel.main(argsraw=[], inputfile=modelpath_3d, get_elemabundances=True)


def test_get_modeldata_1d() -> None:
    for getheadersonly in [False, True]:
        dfmodel, modelmeta = at.get_modeldata(modelpath=modelpath, getheadersonly=getheadersonly)
        assert np.isclose(modelmeta["vmax_cmps"], 800000000.0)
        assert modelmeta["dimensions"] == 1
        assert modelmeta["modelcellcount"] == 1

    dfmodel, modelmeta = at.get_modeldata(modelpath=modelpath)
    assert np.isclose(dfmodel.cellmass_grams.sum(), 1.416963e33)


def test_get_modeldata_3d() -> None:
    for getheadersonly in [False, True]:
        dfmodel, modelmeta = at.get_modeldata(modelpath=modelpath_3d, getheadersonly=getheadersonly)
        assert np.isclose(modelmeta["vmax_cmps"], 2892020000.0)
        assert modelmeta["dimensions"] == 3
        assert modelmeta["modelcellcount"] == 1000
        assert modelmeta["ncoordgrid"] == 10

    dfmodel, modelmeta = at.get_modeldata(modelpath=modelpath_3d)
    assert np.isclose(dfmodel.cellmass_grams.sum(), 2.7861855e33)


def test_downscale_3dmodel() -> None:
    dfmodel, modelmeta = at.get_modeldata(modelpath=modelpath_3d, get_elemabundances=True)
    modelpath_3d_small = at.inputmodel.downscale3dgrid.make_downscaled_3d_grid(modelpath_3d, outputgridsize=2)
    dfmodel_small, modelmeta_small = at.get_modeldata(modelpath_3d_small, get_elemabundances=True)
    assert np.isclose(dfmodel["cellmass_grams"].sum(), dfmodel_small["cellmass_grams"].sum())
    assert np.isclose(modelmeta["vmax_cmps"], modelmeta_small["vmax_cmps"])
    assert np.isclose(modelmeta["t_model_init_days"], modelmeta_small["t_model_init_days"])

    abundcols = (x for x in dfmodel.columns if x.startswith("X_"))
    for abundcol in abundcols:
        assert np.isclose(
            (dfmodel[abundcol] * dfmodel["cellmass_grams"]).sum(),
            (dfmodel_small[abundcol] * dfmodel_small["cellmass_grams"]).sum(),
        )


def test_makemodel_botyanski2017() -> None:
    at.inputmodel.botyanski2017.main(argsraw=[], outputpath=outputpath)


def test_makemodel() -> None:
    at.inputmodel.makeartismodel.main(argsraw=[], modelpath=modelpath, outputpath=outputpath)


def test_makemodel_energyfiles() -> None:
    at.inputmodel.makeartismodel.main(
        argsraw=[], modelpath=modelpath, makeenergyinputfiles=True, modeldim=1, outputpath=outputpath
    )


def test_maketardismodel() -> None:
    at.inputmodel.maketardismodelfromartis.main(argsraw=[], inputpath=modelpath, outputpath=outputpath)


def test_make_empty_abundance_file() -> None:
    at.inputmodel.save_empty_abundance_file(ngrid=50, outputfilepath=outputpath)


def test_opacity_by_Ye_file() -> None:
    griddata = {
        "cellYe": [0, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.5],
        "rho": [0, 99, 99, 99, 99, 99, 99, 99],
        "inputcellid": range(1, 9),
    }
    at.inputmodel.opacityinputfile.opacity_by_Ye(outputpath, griddata=griddata)


def test_save3Dmodel() -> None:
    dfmodel = pd.DataFrame(
        {
            "inputcellid": [1, 2, 3, 4, 5, 6, 7, 8],
            "pos_x_min": [-1, 1, -1, 1, -1, 1, -1, 1],
            "pos_y_min": [-1, -1, 1, 1, -1, -1, 1, 1],
            "pos_z_min": [-1, -1, -1, -1, 1, 1, 1, 1],
            "rho": [0, 2, 3, 2, 5, 7, 8, 2],
            "cellYe": [0, 0.1, 0.2, 0.1, 0.5, 0.1, 0.3, 3],
        }
    )
    tmodel = 100
    vmax = 1000
    at.inputmodel.save_modeldata(
        modelpath=outputpath, dfmodel=dfmodel, t_model_init_days=tmodel, vmax=vmax, dimensions=3
    )



#!/usr/bin/env python3
"""Functions for reading and processing estimator files.

Examples are temperatures, populations, and heating/cooling rates.
"""

from __future__ import annotations

import argparse
import contextlib
import math
import multiprocessing
import sys
import typing as t
from collections import namedtuple
from functools import partial
from functools import reduce
from pathlib import Path

import numpy as np
import pandas as pd
from typeguard import typechecked

import artistools as at


def get_variableunits(key: str | None = None) -> str | dict[str, str]:
    variableunits = {
        "time": "days",
        "gamma_NT": "/s",
        "gamma_R_bfest": "/s",
        "TR": "K",
        "Te": "K",
        "TJ": "K",
        "nne": "e-/cm3",
        "heating": "erg/s/cm3",
        "heating_dep/total_dep": "Ratio",
        "cooling": "erg/s/cm3",
        "velocity": "km/s",
        "velocity_outer": "km/s",
    }
    return variableunits[key] if key else variableunits


def get_variablelongunits(key: str | None = None) -> str | dict[str, str]:
    variablelongunits = {
        "heating_dep/total_dep": "",
        "TR": "Temperature [K]",
        "Te": "Temperature [K]",
        "TJ": "Temperature [K]",
    }
    return variablelongunits[key] if key else variablelongunits


def get_dictlabelreplacements(key: str | None = None) -> str | dict[str, str]:
    dictlabelreplacements = {
        "lognne": "Log nne",
        "Te": "T$_e$",
        "TR": "T$_R$",
        "gamma_NT": r"$\Gamma_{\rm non-thermal}$ [s$^{-1}$]",
        "gamma_R_bfest": r"$\Gamma_{\rm phot}$ [s$^{-1}$]",
        "heating_dep/total_dep": "Heating fraction",
    }
    return dictlabelreplacements[key] if key else dictlabelreplacements


def apply_filters(
    xlist: list[float] | np.ndarray, ylist: list[float] | np.ndarray, args: argparse.Namespace
) -> tuple[list[float] | np.ndarray, list[float] | np.ndarray]:
    filterfunc = at.get_filterfunc(args)

    if filterfunc is not None:
        ylist = filterfunc(ylist)

    return xlist, ylist


def get_ionrecombrates_fromfile(filename: Path | str) -> pd.DataFrame:
    """WARNING: copy pasted from artis-atomic! replace with a package import soon ionstage is the lower ion stage."""
    print(f"Reading {filename}")

    header_row = []
    with Path(filename).open() as filein:
        while True:
            line = filein.readline()
            if line.strip().startswith("TOTAL RECOMBINATION RATE"):
                line = filein.readline()
                line = filein.readline()
                header_row = filein.readline().strip().replace(" n)", "-n)").split()
                break

        if not header_row:
            print("ERROR: no header found")
            sys.exit()

        index_logt = header_row.index("log(T)")
        index_low_n = header_row.index("RRC(low-n)")
        index_tot = header_row.index("RRC(total)")

        recomb_tuple = namedtuple("recomb_tuple", ["logT", "RRC_low_n", "RRC_total"])
        records = []
        for line in filein:
            if row := line.split():
                if len(row) != len(header_row):
                    print("Row contains wrong number of items for header:")
                    print(header_row)
                    print(row)
                    sys.exit()
                records.append(recomb_tuple(*[float(row[index]) for index in [index_logt, index_low_n, index_tot]]))

    return pd.DataFrame.from_records(records, columns=recomb_tuple._fields)


def get_units_string(variable: str) -> str:
    if variable in get_variableunits():
        return f" [{get_variableunits(variable)}]"
    if variable.split("_")[0] in get_variableunits():
        return f' [{get_variableunits(variable.split("_")[0])}]'
    return ""


def parse_estimfile(
    estfilepath: Path,
    modelpath: Path,
    get_ion_values: bool = True,
    get_heatingcooling: bool = True,
    skip_emptycells: bool = False,
) -> t.Iterator[tuple[int, int, dict]]:  # pylint: disable=unused-argument
    """Generate timestep, modelgridindex, dict from estimator file."""
    with at.zopen(estfilepath) as estimfile:
        timestep: int = -1
        modelgridindex: int = -1
        estimblock: dict[t.Any, t.Any] = {}
        for line in estimfile:
            row: list[str] = line.split()
            if not row:
                continue

            if row[0] == "timestep":
                # yield the previous block before starting a new one
                if (
                    timestep >= 0
                    and modelgridindex >= 0
                    and (not skip_emptycells or not estimblock.get("emptycell", True))
                ):
                    yield timestep, modelgridindex, estimblock

                timestep = int(row[1])
                # if timestep > itstep:
                #     print(f"Dropping estimator data from timestep {timestep} and later (> itstep {itstep})")
                #     # itstep in input.txt is updated by ARTIS at every timestep, so the data beyond here
                #     # could be half-written to disk and cause parsing errors
                #     return

                modelgridindex = int(row[3])
                emptycell = row[4] == "EMPTYCELL"
                estimblock = {"emptycell": emptycell}
                if not emptycell:
                    # will be TR, Te, W, TJ, nne
                    for variablename, value in zip(row[4::2], row[5::2]):
                        estimblock[variablename] = float(value)
                    estimblock["lognne"] = math.log10(estimblock["nne"]) if estimblock["nne"] > 0 else float("-inf")

            elif row[1].startswith("Z=") and get_ion_values:
                variablename = row[0]
                if row[1].endswith("="):
                    atomic_number = int(row[2])
                    startindex = 3
                else:
                    atomic_number = int(row[1].split("=")[1])
                    startindex = 2

                estimblock.setdefault(variablename, {})

                for ion_stage_str, value in zip(row[startindex::2], row[startindex + 1 :: 2]):
                    if ion_stage_str.strip() == "(or":
                        continue

                    value_thision = float(value.rstrip(","))

                    if ion_stage_str.strip() == "SUM:":
                        estimblock[variablename][atomic_number] = value_thision
                        continue

                    try:
                        ion_stage = int(ion_stage_str.rstrip(":"))
                    except ValueError:
                        if variablename == "populations" and ion_stage_str.startswith(at.get_elsymbol(atomic_number)):
                            estimblock[variablename][ion_stage_str.rstrip(":")] = float(value)
                        else:
                            print(ion_stage_str, at.get_elsymbol(atomic_number))
                            print(f"Cannot parse row: {row}")
                        continue

                    estimblock[variablename][(atomic_number, ion_stage)] = value_thision

                    if variablename in ["Alpha_R*nne", "AlphaR*nne"]:
                        estimblock.setdefault("Alpha_R", {})
                        estimblock["Alpha_R"][(atomic_number, ion_stage)] = (
                            value_thision / estimblock["nne"] if estimblock["nne"] > 0.0 else float("inf")
                        )

                    else:  # variablename == 'populations':
                        # contribute the ion population to the element population
                        estimblock[variablename].setdefault(atomic_number, 0.0)
                        estimblock[variablename][atomic_number] += value_thision

                if variablename == "populations":
                    # contribute the element population to the total population
                    estimblock["populations"].setdefault("total", 0.0)
                    estimblock["populations"]["total"] += estimblock["populations"][atomic_number]
                    estimblock.setdefault("nntot", 0.0)
                    estimblock["nntot"] += estimblock["populations"][atomic_number]

            elif row[0] == "heating:" and get_heatingcooling:
                for heatingtype, value in zip(row[1::2], row[2::2]):
                    key = heatingtype if heatingtype.startswith("heating_") else "heating_" + heatingtype
                    estimblock[key] = float(value)

                if "heating_gamma/gamma_dep" in estimblock and estimblock["heating_gamma/gamma_dep"] > 0:
                    estimblock["gamma_dep"] = estimblock["heating_gamma"] / estimblock["heating_gamma/gamma_dep"]
                elif "heating_dep/total_dep" in estimblock and estimblock["heating_dep/total_dep"] > 0:
                    estimblock["total_dep"] = estimblock["heating_dep"] / estimblock["heating_dep/total_dep"]

            elif row[0] == "cooling:" and get_heatingcooling:
                for coolingtype, value in zip(row[1::2], row[2::2]):
                    estimblock["cooling_" + coolingtype] = float(value)

    # reached the end of file
    if timestep >= 0 and modelgridindex >= 0 and (not skip_emptycells or not estimblock.get("emptycell", True)):
        yield timestep, modelgridindex, estimblock


def read_estimators_from_file(
    folderpath: Path | str,
    modelpath: Path,
    arr_velocity_outer: t.Sequence[float] | None,
    mpirank: int,
    printfilename: bool = False,
    get_ion_values: bool = True,
    get_heatingcooling: bool = True,
    skip_emptycells: bool = False,
) -> dict[tuple[int, int], t.Any]:
    estimators_thisfile = {}
    estimfilename = f"estimators_{mpirank:04d}.out"
    try:
        estfilepath = at.firstexisting(estimfilename, folder=folderpath, tryzipped=True)
    except FileNotFoundError:
        # not worth printing an error, because ranks with no cells to update do not produce an estimator file
        # print(f'Warning: Could not find {estfilepath.relative_to(modelpath.parent)}')
        return {}

    if printfilename:
        filesize = Path(estfilepath).stat().st_size / 1024 / 1024
        print(f"Reading {estfilepath.relative_to(modelpath.parent)} ({filesize:.2f} MiB)")

    for fileblock_timestep, fileblock_modelgridindex, file_estimblock in parse_estimfile(
        estfilepath,
        modelpath,
        get_ion_values=get_ion_values,
        get_heatingcooling=get_heatingcooling,
        skip_emptycells=skip_emptycells,
    ):
        if arr_velocity_outer is not None:
            file_estimblock["velocity_outer"] = arr_velocity_outer[fileblock_modelgridindex]
            file_estimblock["velocity"] = file_estimblock["velocity_outer"]

        estimators_thisfile[(fileblock_timestep, fileblock_modelgridindex)] = file_estimblock

    return estimators_thisfile


@typechecked
def read_estimators(
    modelpath: Path | str = Path(),
    modelgridindex: None | int | t.Sequence[int] = None,
    timestep: None | int | t.Sequence[int] = None,
    mpirank: int | None = None,
    runfolder: None | str | Path = None,
    get_ion_values: bool = True,
    get_heatingcooling: bool = True,
    skip_emptycells: bool = False,
    add_velocity: bool = True,
) -> dict[tuple[int, int], dict]:
    """Read estimator files into a nested dictionary structure.

    Speed it up by only retrieving estimators for a particular timestep(s) or modelgrid cells.
    """
    modelpath = Path(modelpath)
    match_modelgridindex: t.Collection[int]
    if modelgridindex is None:
        match_modelgridindex = []
    elif isinstance(modelgridindex, int):
        match_modelgridindex = (modelgridindex,)
    else:
        match_modelgridindex = tuple(modelgridindex)

    if -1 in match_modelgridindex:
        match_modelgridindex = []

    match_timestep: t.Collection[int]
    if timestep is None:
        match_timestep = []
    elif isinstance(timestep, int):
        match_timestep = (timestep,)
    else:
        match_timestep = tuple(timestep)

    if not Path(modelpath).exists() and Path(modelpath).parts[0] == "codecomparison":
        return at.codecomparison.read_reference_estimators(modelpath, timestep=timestep, modelgridindex=modelgridindex)

    # print(f" matching cells {match_modelgridindex} and timesteps {match_timestep}")

    arr_velocity_outer = None
    if add_velocity:
        modeldata, _ = at.inputmodel.get_modeldata(modelpath, getheadersonly=True)
        if "velocity_outer" in modeldata.columns:
            modeldata, _ = at.inputmodel.get_modeldata(modelpath)
            arr_velocity_outer = tuple(float(v) for v in modeldata["velocity_outer"].to_numpy())

    mpiranklist = (
        at.get_mpiranklist(modelpath, modelgridindex=match_modelgridindex, only_ranks_withgridcells=True)
        if mpirank is None
        else [mpirank]
    )

    runfolders = at.get_runfolders(modelpath, timesteps=match_timestep) if runfolder is None else [Path(runfolder)]

    printfilename = len(mpiranklist) < 10

    estimators: dict[tuple[int, int], dict] = {}
    for folderpath in runfolders:
        if not printfilename:
            print(
                f"Reading {len(list(mpiranklist))} estimator files in {folderpath.relative_to(Path(modelpath).parent)}"
            )

        processfile = partial(
            read_estimators_from_file,
            folderpath,
            modelpath,
            arr_velocity_outer,
            get_ion_values=get_ion_values,
            get_heatingcooling=get_heatingcooling,
            printfilename=printfilename,
            skip_emptycells=skip_emptycells,
        )

        if at.get_config()["num_processes"] > 1:
            with multiprocessing.get_context("spawn").Pool(processes=at.get_config()["num_processes"]) as pool:
                arr_rankestimators = pool.map(processfile, mpiranklist)
                pool.close()
                pool.join()
                pool.terminate()
        else:
            arr_rankestimators = [processfile(rank) for rank in mpiranklist]

        for mpirank, estimators_thisfile in zip(mpiranklist, arr_rankestimators):
            dupekeys = sorted([k for k in estimators_thisfile if k in estimators])
            for k in dupekeys:
                # dropping the lowest timestep is normal for restarts. Only warn about other cases
                if k[0] != dupekeys[0][0]:
                    filepath = Path(folderpath, f"estimators_{mpirank:04d}.out")
                    print(
                        f"WARNING: Duplicate estimator block for (timestep, mgi) key {k}. "
                        f"Dropping block from {filepath}"
                    )

                del estimators_thisfile[k]

            estimators |= estimators_thisfile

    return estimators


def get_averaged_estimators(
    modelpath: Path | str,
    estimators: dict[tuple[int, int], dict],
    timesteps: int | t.Sequence[int],
    modelgridindex: int,
    keys: str | list,
    avgadjcells: int = 0,
) -> t.Any | float:
    """Get the average of estimators[(timestep, modelgridindex)][keys[0]]...[keys[-1]] across timesteps."""
    if isinstance(keys, str):
        keys = [keys]

    # reduce(lambda d, k: d[k], keys, dictionary) returns dictionary[keys[0]][keys[1]]...[keys[-1]]
    # applying all keys in the keys list

    # if single timestep, no averaging needed
    if isinstance(timesteps, int):
        return reduce(lambda d, k: d[k], [(timesteps, modelgridindex), *keys], estimators)

    firsttimestepvalue = reduce(lambda d, k: d[k], [(timesteps[0], modelgridindex), *keys], estimators)
    if isinstance(firsttimestepvalue, dict):
        return {
            k: get_averaged_estimators(modelpath, estimators, timesteps, modelgridindex, [*keys, k])
            for k in firsttimestepvalue
        }

    tdeltas = at.get_timestep_times(modelpath, loc="delta")
    valuesum = 0
    tdeltasum = 0
    for timestep, tdelta in zip(timesteps, tdeltas):
        for mgi in range(modelgridindex - avgadjcells, modelgridindex + avgadjcells + 1):
            with contextlib.suppress(KeyError):
                valuesum += reduce(lambda d, k: d[k], [(timestep, mgi), *keys], estimators) * tdelta
                tdeltasum += tdelta
    return valuesum / tdeltasum

    # except KeyError:
    #     if (timestep, modelgridindex) in estimators:
    #         print(f'Unknown x variable: {xvariable} for timestep {timestep} in cell {modelgridindex}')
    #     else:
    #         print(f'No data for cell {modelgridindex} at timestep {timestep}')
    #     print(estimators[(timestep, modelgridindex)])
    #     sys.exit()


def get_averageionisation(populations: dict[t.Any, float], atomic_number: int) -> float:
    free_electron_weighted_pop_sum = 0.0
    found = False
    popsum = 0.0
    for key in populations:
        if isinstance(key, tuple) and key[0] == atomic_number:
            found = True
            ion_stage = key[1]
            free_electron_weighted_pop_sum += populations[key] * (ion_stage - 1)
            popsum += populations[key]

    if not found:
        return float("NaN")

    return free_electron_weighted_pop_sum / populations[atomic_number]


def get_averageexcitation(
    modelpath: Path, modelgridindex: int, timestep: int, atomic_number: int, ion_stage: int, T_exc: float
) -> float:
    dfnltepops = at.nltepops.read_files(modelpath, modelgridindex=modelgridindex, timestep=timestep)
    adata = at.atomic.get_levels(modelpath)
    ionlevels = adata.query("Z == @atomic_number and ion_stage == @ion_stage").iloc[0].levels

    energypopsum = 0
    ionpopsum = 0
    if dfnltepops.empty:
        return float("NaN")

    dfnltepops_ion = dfnltepops.query(
        "modelgridindex==@modelgridindex and timestep==@timestep and Z==@atomic_number & ion_stage==@ion_stage"
    )

    k_b = 8.617333262145179e-05  # eV / K

    ionpopsum = dfnltepops_ion.n_NLTE.sum()
    energypopsum = (
        dfnltepops_ion[dfnltepops_ion.level >= 0].eval("@ionlevels.iloc[level].energy_ev.values * n_NLTE").sum()
    )

    with contextlib.suppress(IndexError):  # no superlevel with cause IndexError
        superlevelrow = dfnltepops_ion[dfnltepops_ion.level < 0].iloc[0]
        levelnumber_sl = dfnltepops_ion.level.max() + 1

        energy_boltzfac_sum = (
            ionlevels.iloc[levelnumber_sl:].eval("energy_ev * g * exp(- energy_ev / @k_b / @T_exc)").sum()
        )

        boltzfac_sum = ionlevels.iloc[levelnumber_sl:].eval("g * exp(- energy_ev / @k_b / @T_exc)").sum()
        # adjust to the actual superlevel population from ARTIS
        energypopsum += energy_boltzfac_sum * superlevelrow.n_NLTE / boltzfac_sum
    return energypopsum / ionpopsum


def get_partiallycompletetimesteps(estimators: dict[tuple[int, int], dict]) -> list[int]:
    """During a simulation, some estimator files can contain information for some cells but not others
    for the current timestep.
    """
    timestepcells: dict[int, list[int]] = {}
    all_mgis = set()
    for nts, mgi in estimators:
        if nts not in timestepcells:
            timestepcells[nts] = []
        timestepcells[nts].append(mgi)
        all_mgis.add(mgi)

    return [nts for nts, mgilist in timestepcells.items() if len(mgilist) < len(all_mgis)]



# PYTHON_ARGCOMPLETE_OK
from __future__ import annotations

import argparse
import importlib

import argcomplete

from .commands import dictcommands as atdictcommands
from .misc import CustomArgHelpFormatter


def addargs(parser: argparse.ArgumentParser) -> None:
    pass


def addsubparsers(parser, parentcommand, dictcommands, depth: int = 1) -> None:
    def func(args) -> None:
        parser.print_help()

    parser.set_defaults(func=func)
    subparsers = parser.add_subparsers(dest=f"{parentcommand} command", required=False)

    for subcommand, subcommands in dictcommands.items():
        strhelp: str | None
        if isinstance(subcommands, dict):
            strhelp = "command group"
            submodule = None
        else:
            submodulename, funcname = subcommands
            namestr = f"artistools.{submodulename.removeprefix('artistools.')}" if submodulename else "artistools"
            submodule = importlib.import_module(namestr, package="artistools")
            func = getattr(submodule, funcname)
            strhelp = func.__doc__

        subparser = subparsers.add_parser(subcommand, help=strhelp, formatter_class=CustomArgHelpFormatter)

        if submodule:
            submodule.addargs(subparser)
            subparser.set_defaults(func=func)
        else:
            addsubparsers(parser=subparser, parentcommand=subcommand, dictcommands=subcommands, depth=depth + 1)


def main(args=None, argsraw=None, **kwargs) -> None:
    """Parse and run artistools commands."""
    parser = argparse.ArgumentParser(
        formatter_class=CustomArgHelpFormatter,
        description="Artistools base command.",
    )
    parser.set_defaults(func=None)

    addsubparsers(parser, "artistools", atdictcommands)

    argcomplete.autocomplete(parser)
    args = parser.parse_args(argsraw)
    args.func(args=args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
from __future__ import annotations

import argparse
import math
import typing as t
from pathlib import Path

import argcomplete
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u

import artistools as at

if t.TYPE_CHECKING:
    import pandas as pd


def plot_2d_initial_abundances(modelpath: Path, args: argparse.Namespace) -> None:
    model = at.inputmodel.get_2d_modeldata(modelpath)
    abundances = at.inputmodel.get_initelemabundances(modelpath)

    abundances["inputcellid"] = abundances["inputcellid"].apply(float)

    merge_dfs = model.merge(abundances, how="inner", on="inputcellid")

    with Path(modelpath, "model.txt").open() as fmodelin:
        fmodelin.readline()  # npts r, npts z
        t_model = float(fmodelin.readline())  # days
        vmax = float(fmodelin.readline())  # v_max in [cm/s]

    r = merge_dfs["cellpos_mid[r]"] / t_model * (u.cm / u.day).to("km/s") / 10**3
    z = merge_dfs["cellpos_mid[z]"] / t_model * (u.cm / u.day).to("km/s") / 10**3

    colname = f"X_{args.plotvars}"
    font = {"weight": "bold", "size": 18}

    f = plt.figure(figsize=(4, 5))
    ax = f.add_subplot(111)
    im = ax.scatter(r, z, c=merge_dfs[colname], marker="8")

    f.colorbar(im)
    plt.xlabel(r"v$_x$ in 10$^3$ km/s", fontsize="x-large")  # , fontweight='bold')
    plt.ylabel(r"v$_z$ in 10$^3$ km/s", fontsize="x-large")  # , fontweight='bold')
    plt.text(20, 25, args.plotvars, color="white", fontweight="bold", fontsize="x-large")
    plt.tight_layout()
    # ax.labelsize: 'large'
    # plt.title(f'At {sliceaxis} = {sliceposition}')

    outfilename = f"plotcomposition{args.plotvars}.pdf"
    plt.savefig(Path(modelpath) / outfilename, format="pdf")
    print(f"Saved {outfilename}")


def get_2D_slice_through_3d_model(
    dfmodel: pd.DataFrame,
    sliceaxis: t.Literal["x", "y", "z"],
    modelmeta: dict[str, t.Any] | None = None,
    plotaxis1: t.Literal["x", "y", "z"] | None = None,
    plotaxis2: t.Literal["x", "y", "z"] | None = None,
    sliceindex: int | None = None,
) -> pd.DataFrame:
    if not sliceindex:
        # get midpoint
        sliceposition: float = dfmodel.iloc[(dfmodel["pos_x_min"]).abs().argsort()][:1]["pos_x_min"].item()
        # Choose position to slice. This gets minimum absolute value as the closest to 0
    else:
        cell_boundaries = []
        for x in dfmodel[f"pos_{sliceaxis}_min"]:
            if x not in cell_boundaries:
                cell_boundaries.append(x)
        sliceposition = cell_boundaries[sliceindex]

    slicedf = dfmodel.loc[dfmodel[f"pos_{sliceaxis}_min"] == sliceposition]

    if modelmeta is not None and plotaxis1 is not None and plotaxis2 is not None:
        assert slicedf.shape[0] == modelmeta[f"ncoordgrid{plotaxis1}"] * modelmeta[f"ncoordgrid{plotaxis2}"]

    return slicedf


def plot_slice_modelcol(ax, dfmodelslice, modelmeta, colname, plotaxis1, plotaxis2, t_model_d, args):
    print(f"plotting {colname}")
    colorscale = (
        dfmodelslice[colname] * dfmodelslice["rho"] if colname.startswith("X_") else dfmodelslice[colname]
    ).to_numpy()

    if args.hideemptycells:
        # Don't plot empty cells:
        colorscale = np.ma.masked_where(colorscale == 0.0, colorscale)

    if args.logcolorscale:
        # logscale for colormap
        with np.errstate(divide="ignore"):
            colorscale = np.log10(colorscale)

    normalise_between_0_and_1 = False
    if normalise_between_0_and_1:
        norm = mpl.colors.Normalize(vmin=0, vmax=1)
        scaledmap = mpl.cm.ScalarMappable(cmap="viridis", norm=norm)
        scaledmap.set_array([])
        colorscale = scaledmap.to_rgba(colorscale)  # colorscale fixed between 0 and 1
    else:
        scaledmap = None

    cmps_to_kmps = 1e-5
    cmps_to_beta = 1.0 / (2.99792458e10)
    unitfactor = cmps_to_beta
    t_model_s = t_model_d * 86400.0

    # take flat array and turn in into 2D array
    grid = colorscale.reshape((modelmeta[f"ncoordgrid{plotaxis1}"], modelmeta[f"ncoordgrid{plotaxis2}"]))

    im = ax.imshow(
        grid,
        cmap="viridis",
        interpolation="nearest",
        extent=(
            dfmodelslice[f"pos_{plotaxis1}_min"].min() / t_model_s * unitfactor,
            dfmodelslice[f"pos_{plotaxis1}_max"].max() / t_model_s * unitfactor,
            dfmodelslice[f"pos_{plotaxis2}_min"].min() / t_model_s * unitfactor,
            dfmodelslice[f"pos_{plotaxis2}_max"].max() / t_model_s * unitfactor,
        ),
        origin="lower",
        # vmin=0.0,
        # vmax=1.0,
    )

    plot_vmax = 0.2
    ax.set_ylim(bottom=-plot_vmax, top=plot_vmax)
    ax.set_xlim(left=-plot_vmax, right=plot_vmax)
    if "_" in colname:
        ax.annotate(
            colname.split("_")[1],
            color="white",
            xy=(0.9, 0.9),
            xycoords="axes fraction",
            horizontalalignment="right",
            verticalalignment="top",
            # fontsize=10,
        )

    return im, scaledmap


def plot_3d_initial_abundances(modelpath, args=None) -> None:
    font = {
        # 'weight': 'bold',
        "size": 18
    }
    mpl.rc("font", **font)

    dfmodel, modelmeta = at.get_modeldata(
        modelpath,
        skipnuclidemassfraccolumns=True,
        get_elemabundances=True,
        dtype_backend="pyarrow",
        derived_cols=["pos_max"],
        use_polars=False,
    )

    targetmodeltime_days = None
    if targetmodeltime_days is not None:
        print(
            f"Scaling modeldata to {targetmodeltime_days} days. \nWARNING: abundances not scaled for radioactive decays"
        )

        dfmodel, modelmeta = at.inputmodel.scale_model_to_time(
            targetmodeltime_days=targetmodeltime_days, modelmeta=modelmeta, dfmodel=dfmodel
        )

    sliceaxis: t.Literal["x", "y", "z"] = "z"

    axes: list[t.Literal["x", "y", "z"]] = ["x", "y", "z"]
    plotaxis1: t.Literal["x", "y", "z"] = next(ax for ax in axes if ax != sliceaxis)
    plotaxis2: t.Literal["x", "y", "z"] = next(ax for ax in axes if ax not in [sliceaxis, plotaxis1])

    df2dslice = get_2D_slice_through_3d_model(
        dfmodel=dfmodel, modelmeta=modelmeta, sliceaxis=sliceaxis, plotaxis1=plotaxis1, plotaxis2=plotaxis2
    )

    subplots = bool(len(args.plotvars) > 1)
    if not subplots:
        fig, ax = plt.subplots(1, 1, figsize=(8, 7), tight_layout={"pad": 0.4, "w_pad": 0.0, "h_pad": 0.0})
    else:
        rows = 1
        cols = len(args.plotvars)

        fig, axes = plt.subplots(
            nrows=rows,
            ncols=cols,
            sharex=True,
            sharey=True,
            figsize=(at.get_config()["figwidth"] * cols, at.get_config()["figwidth"] * 1.4),
            # tight_layout={"pad": 5.0, "w_pad": 0.0, "h_pad": 0.0},
        )
        for ax in axes:
            ax.set(aspect="equal")

    for index, plotvar in enumerate(args.plotvars):
        colname = plotvar if plotvar in df2dslice.columns else f"X_{plotvar}"

        if subplots:
            ax = axes[index]
        im, scaledmap = plot_slice_modelcol(
            ax, df2dslice, modelmeta, colname, plotaxis1, plotaxis2, modelmeta["t_model_init_days"], args
        )

    xlabel = rf"v$_{plotaxis1}$ [$c$]"
    ylabel = rf"v$_{plotaxis2}$ [$c$]"

    if not subplots:
        cbar = fig.colorbar(im)
        plt.xlabel(xlabel, fontsize="x-large")  # , fontweight='bold')
        plt.ylabel(ylabel, fontsize="x-large")  # , fontweight='bold')
    else:
        cbar = fig.colorbar(scaledmap, ax=axes, shrink=cols * 0.08, location="top", pad=0.8, anchor=(0.5, 3.0))
        fig.text(0.5, 0.15, xlabel, ha="center", va="center")
        fig.text(0.05, 0.5, ylabel, ha="center", va="center", rotation="vertical")

    if "cellYe" not in args.plotvars and "tracercount" not in args.plotvars:
        if args.logcolorscale:
            cbar.ax.set_title(r"log10($\rho$) [g/cm3]", size="small")
        else:
            cbar.ax.set_title(r"$\rho$ [g/cm3]", size="small")
    # cbar.ax.tick_params(labelsize='x-large')

    # plt.tight_layout()
    # ax.labelsize: 'large'
    # plt.title(f'At {sliceaxis} = {sliceposition}')

    outfilename = args.outputfile or f"plotcomposition_{','.join(args.plotvars)}.pdf"
    plt.savefig(Path(modelpath) / outfilename, format="pdf")

    print(f"Saved {outfilename}")


def get_model_abundances_Msun_1D(modelpath: Path) -> pd.DataFrame:
    filename = modelpath / "model.txt"
    modeldata, t_model_init_days, _ = at.inputmodel.get_modeldata_tuple(filename)
    abundancedata = at.inputmodel.get_initelemabundances(modelpath)

    t_model_init_seconds = t_model_init_days * 24 * 60 * 60

    modeldata["volume_shell"] = (
        4
        / 3
        * math.pi
        * (
            (modeldata["velocity_outer"] * 1e5 * t_model_init_seconds) ** 3
            - (modeldata["velocity_inner"] * 1e5 * t_model_init_seconds) ** 3
        )
    )

    modeldata["mass_shell"] = (10 ** modeldata["logrho"]) * modeldata["volume_shell"]

    merge_dfs = modeldata.merge(abundancedata, how="inner", on="inputcellid")

    print("Total mass (Msun):")
    for key in merge_dfs:
        if "X_" in key:
            merge_dfs[f"mass_{key}"] = merge_dfs[key] * merge_dfs["mass_shell"] * u.g.to("solMass")
            # get mass of element in each cell
            print(key, merge_dfs[f"mass_{key}"].sum())  # print total mass of element in solmass

    return merge_dfs


def plot_most_abundant(modelpath, args):
    model, _ = at.inputmodel.get_modeldata(modelpath[0])
    abundances = at.inputmodel.get_initelemabundances(modelpath[0])

    merge_dfs = model.merge(abundances, how="inner", on="inputcellid")
    elements = [x for x in merge_dfs if "X_" in x]

    merge_dfs["max"] = merge_dfs[elements].idxmax(axis=1)

    merge_dfs["max"] = merge_dfs["max"].apply(lambda x: at.get_atomic_number(x[2:]))
    return merge_dfs[merge_dfs["max"] != 1]


def make_3d_plot(modelpath, args):
    import pyvista as pv

    pv.set_plot_theme("document")  # set white background

    model, t_model, vmax = at.inputmodel.get_modeldata_tuple(modelpath, get_elemabundances=False)
    abundances = at.inputmodel.get_initelemabundances(modelpath)

    abundances["inputcellid"] = abundances["inputcellid"].apply(float)

    merge_dfs = model.merge(abundances, how="inner", on="inputcellid")
    model = merge_dfs

    # choose what surface will be coloured by
    if args.rho:
        coloursurfaceby = "rho"
    elif args.opacity:
        model["opacity"] = at.inputmodel.opacityinputfile.get_opacity_from_file(modelpath)
        coloursurfaceby = "opacity"
    else:
        print(f"Colours set by X_{args.plotvars}")
        coloursurfaceby = f"X_{args.plotvars}"

    # generate grid from data
    grid = round(len(model["rho"]) ** (1.0 / 3.0))
    surfacecolorscale = np.zeros((grid, grid, grid))  # needs 3D array
    xgrid = np.zeros(grid)

    surfacearr = np.array(model[coloursurfaceby])

    i = 0
    for z in range(grid):
        for y in range(grid):
            for x in range(grid):
                surfacecolorscale[x, y, z] = surfacearr[i]
                xgrid[x] = -vmax + 2 * x * vmax / grid
                i += 1

    x, y, z = np.meshgrid(xgrid, xgrid, xgrid)

    mesh = pv.StructuredGrid(x, y, z)
    print(mesh)  # tells you the properties of the mesh

    mesh[coloursurfaceby] = surfacecolorscale.ravel(order="F")  # add data to the mesh
    # mesh.plot()
    minval = np.min(mesh[coloursurfaceby][np.nonzero(mesh[coloursurfaceby])])  # minimum non zero value
    print(f"{coloursurfaceby} minumin {minval}, maximum {max(mesh[coloursurfaceby])}")

    if not args.surfaces3d:
        surfacepositions = np.linspace(min(mesh[coloursurfaceby]), max(mesh[coloursurfaceby]), num=10)
        print(f"Using default surfaces {surfacepositions} \n define these with -surfaces3d for better results")
    else:
        surfacepositions = args.surfaces3d
    # surfacepositions = [1, 50, 100, 300, 500, 800, 1000, 1100, 1200, 1300, 1400, 1450, 1500] # choose these

    surf = mesh.contour(surfacepositions, scalars=coloursurfaceby)  # create isosurfaces

    surf.plot(opacity="linear", screenshot=modelpath / "3Dplot.png")  # plot surfaces and save screenshot


def addargs(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-modelpath",
        type=Path,
        default=Path(),
        help="Path to ARTIS folder",
    )

    parser.add_argument("-o", action="store", dest="outputfile", type=Path, default=None, help="Filename for PDF file")

    parser.add_argument(
        "plotvars",
        type=str,
        default=["rho"],
        nargs="+",
        help=(
            "Element symbols (Fe, Ni, Sr) for mass fraction or other model columns (rho, tracercount) to plot. Default"
            " is rho"
        ),
    )

    parser.add_argument("--logcolorscale", action="store_true", help="Use log scale for colour map")

    parser.add_argument("--hideemptycells", action="store_true", help="Don't plot empty cells")

    parser.add_argument("--opacity", action="store_true", help="Plot opacity from opacity.txt (if available for model)")

    parser.add_argument("--plot3d", action="store_true", help="Make 3D plot")

    parser.add_argument("-surfaces3d", type=float, nargs="+", help="define positions of surfaces for 3D plots")


def main(args=None, argsraw=None, **kwargs) -> None:
    """Plot ARTIS input model composition."""
    if args is None:
        parser = argparse.ArgumentParser(formatter_class=at.CustomArgHelpFormatter, description=__doc__)
        addargs(parser)
        parser.set_defaults(**kwargs)
        argcomplete.autocomplete(parser)
        args = parser.parse_args(argsraw)

    if not args.modelpath:
        args.modelpath = ["."]

    if args.plot3d:
        make_3d_plot(Path(args.modelpath), args)
        return

    _, modelmeta = at.get_modeldata(modelpath=args.modelpath, getheadersonly=True, printwarningsonly=True)

    if modelmeta["dimensions"] == 2:
        plot_2d_initial_abundances(args.modelpath, args)

    elif modelmeta["dimensions"] == 3:
        plot_3d_initial_abundances(args.modelpath, args)


if __name__ == "__main__":
    main()



"""artistools.

A collection of plotting, analysis, and file format conversion tools
for the ARTIS radiative transfer code.
"""

import typing as t

from . import atomic
from . import codecomparison
from . import commands
from . import deposition
from . import estimators
from . import initial_composition
from . import inputmodel
from . import lightcurve
from . import macroatom
from . import nltepops
from . import nonthermal
from . import packets
from . import plotspherical
from . import radfield
from . import spectra
from . import transitions
from . import writecomparisondata
from .__main__ import addargs
from .__main__ import main
from .configuration import get_config
from .configuration import set_config
from .inputmodel import add_derived_cols_to_modeldata
from .inputmodel import get_2d_modeldata
from .inputmodel import get_cell_angle
from .inputmodel import get_dfmodel_dimensions
from .inputmodel import get_mean_cell_properties_of_angle_bin
from .inputmodel import get_mgi_of_velocity_kms
from .inputmodel import get_modeldata
from .inputmodel import get_modeldata_tuple
from .inputmodel import save_initelemabundances
from .inputmodel import save_modeldata
from .misc import anyexist
from .misc import AppendPath
from .misc import average_direction_bins
from .misc import CustomArgHelpFormatter
from .misc import decode_roman_numeral
from .misc import firstexisting
from .misc import flatten_list
from .misc import get_atomic_number
from .misc import get_bflist
from .misc import get_cellsofmpirank
from .misc import get_composition_data
from .misc import get_composition_data_from_outputfile
from .misc import get_costheta_bins
from .misc import get_costhetabin_phibin_labels
from .misc import get_deposition
from .misc import get_dirbin_labels
from .misc import get_elsymbol
from .misc import get_elsymbolslist
from .misc import get_escaped_arrivalrange
from .misc import get_file_metadata
from .misc import get_filterfunc
from .misc import get_grid_mapping
from .misc import get_inputparams
from .misc import get_ionstring
from .misc import get_linelist_dataframe
from .misc import get_linelist_dict
from .misc import get_linelist_pldf
from .misc import get_model_name
from .misc import get_mpiranklist
from .misc import get_mpirankofcell
from .misc import get_nprocs
from .misc import get_nu_grid
from .misc import get_phi_bins
from .misc import get_runfolders
from .misc import get_syn_dir
from .misc import get_time_range
from .misc import get_timestep_of_timedays
from .misc import get_timestep_time
from .misc import get_timestep_times
from .misc import get_viewingdirection_costhetabincount
from .misc import get_viewingdirection_phibincount
from .misc import get_viewingdirectionbincount
from .misc import get_vpkt_config
from .misc import get_vspec_dir_labels
from .misc import get_wid_init_at_tmin
from .misc import get_wid_init_at_tmodel
from .misc import get_z_a_nucname
from .misc import join_pdf_files
from .misc import linetuple
from .misc import makelist
from .misc import match_closest_time
from .misc import namedtuple
from .misc import parse_range
from .misc import parse_range_list
from .misc import read_linestatfile
from .misc import readnoncommentline
from .misc import roman_numerals
from .misc import showtimesteptimes
from .misc import split_dataframe_dirbins
from .misc import stripallsuffixes
from .misc import trim_or_pad
from .misc import vec_len
from .misc import zopen
from .plottools import set_mpl_style


def get_path(**kwargs: t.Any) -> None:
    print(get_config("path_artistools_dir"))


set_mpl_style()
